from __future__ import annotations

import shutil
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path

from pydantic import BaseModel


def recursive_mkdir(project_path: Path, structure: dict | list | Path) -> None:
    """Recursively make the provided directory structure."""
    if isinstance(structure, dict):
        for key, value in structure.items():
            Path(project_path, key).mkdir(exist_ok=True, parents=True)
            recursive_mkdir(Path(project_path, key), value)
    elif isinstance(structure, list):
        for value in structure:
            Path(project_path, value).mkdir(exist_ok=True, parents=True)


class BaseProject(BaseModel, ABC):
    name: str
    project_path: Path
    venv_path: Path
    dir_structure: dict[str, dict | list | Path] | list[Path] | Path

    _engine = ""

    @property
    def engine(self):
        return self._engine

    @abstractmethod
    def create(self, python_path: Path):
        """Create the project."""
        Path(self.project_path, ".aiida").mkdir(parents=True, exist_ok=True)
        recursive_mkdir(self.project_path, self.dir_structure)

    @abstractmethod
    def destroy(self):
        """Destroy the project."""
        shutil.rmtree(self.project_path, ignore_errors=True)

    @abstractmethod
    def append_activate_text(self, text: str) -> None:
        """Append a text to the activate script."""

    @abstractmethod
    def append_deactivate_text(self, text: str) -> None:
        """Append a text to the deactivate script."""

    @abstractmethod
    def install(self, package):
        """Install a package from PyPI."""

    @abstractmethod
    def install_local(self, path):
        """Install a package from a local directory."""

    def clone_repo(self, repo, path):
        """Clone a ``git`` repository from a remote source."""
        clone_command = ["git", "clone", "--single-branch", f"{repo}", f"{path.resolve()}"]
        subprocess.run(clone_command, capture_output=True)



#!/bin/bash
"""Script for updating the `CHANGELOG.md` based on the commits since the latest release tag."""
import re
import subprocess
from pathlib import Path

DEFAULT_CHANGELOG_SECTIONS = """
### ‼️ Breaking changes


### ✨ New features


### 🗑️ Deprecations


### 👌 Improvements


### 🐛 Bug fixes


### 📚 Documentation


### 🔧 Maintenance


### ⬆️ Update dependencies


### ♻️ Refactor

"""


def update_changelog():
    """Update the `CHANGELOG.md` for a first draft of the release."""

    print("🔍 Checking the current version number")
    current_changelog = Path("CHANGELOG.md").read_text(encoding="utf-8")

    from aiida_project import __version__

    if str(__version__) in current_changelog:
        print("🛑 Current version already in `CHANGELOG.md`. Skipping...")
        return

    print("⬆️ Found updated version number, adapting `CHANGELOG.md`.")
    tags = subprocess.run(
        ["git", "tag", "--sort=v:refname"], capture_output=True, check=True, encoding="utf-8"
    ).stdout
    latest_tag = re.findall(r"(v\d\.\d\.\d)\n", tags)[-1]

    print(f"🔄 Comparing with latest tag `{latest_tag}`.")
    commits = subprocess.run(
        ["git", "log", "--pretty=format:'%h|%H|%s'", f"{latest_tag}..origin/main"],
        capture_output=True,
        check=True,
        encoding="utf-8",
    ).stdout

    pr_pattern = re.compile(r"\(\S(?P<pr_number>\d+)\)")

    changelog_message = f"## v{__version__}\n" + DEFAULT_CHANGELOG_SECTIONS

    for commit in commits.splitlines():
        # Remove the PR number from the commit message
        pr_match = pr_pattern.search(commit)

        if pr_match is not None:
            pr_number = pr_match.groupdict()["pr_number"]
            commit = commit.replace(rf"(#{pr_number})", "")

        # Add the commit hash (short) to link to the changelog
        commit = commit.strip("'")
        hash_short, hash_long, message = commit.split("|", maxsplit=2)
        message += (
            f" [[{hash_short}](https://github.com/aiidateam/aiida-project/commit/{hash_long})]"
        )
        changelog_message += f"\n* {message}"

    with Path("CHANGELOG.md").open("w", encoding="utf8") as handle:
        handle.write(changelog_message + "\n\n" + current_changelog)

    print("🚀 Success! Finalise the `CHANGELOG.md` and let's get this baby released.")


if __name__ == "__main__":
    update_changelog()



import os
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import List, Optional

import typer
from rich import print, prompt
from typing_extensions import Annotated

from ..project import EngineType, load_project_class
from ..shell import ShellType, load_shell

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.callback()
def callback():
    """
    Tool for importing CIF files and converting them into a unique set of `StructureData`.
    """


@app.command()
def init(shell: Optional[ShellType] = None):
    """Initialisation of the `aiida-project` setup."""
    from ..config import ProjectConfig

    config = ProjectConfig()

    shell_str = shell.value if shell else ""

    if not shell_str:
        shell_guess = os.environ.get("SHELL", "")
        detected_shell = shell_guess.split("/")[-1] if shell_guess else None
        prompt_message = "👋 Hello there! Which shell are you using?"
        shell_str = prompt.Prompt.ask(
            prompt=prompt_message,
            choices=[shell_type.value for shell_type in ShellType],
            default=detected_shell,
        )

    config.set_key("aiida_project_shell", shell_str)
    shellz = load_shell(shell_str)

    shellz.config_file.touch(exist_ok=True)

    if "Created by `aiida-project init`" in shellz.config_file.read_text():
        print(
            "[bold blue]Report:[/] There is already an `aiida-project` initialization comment in "
            f"{shellz.config_file}"
        )
        add_init_lines = prompt.Confirm.ask("Do you want still want to add the init lines?")
    else:
        add_init_lines = True

    if add_init_lines:
        with shellz.config_file.open("a") as handle:
            handle.write(
                f"\n# Created by `aiida-project init` on "
                f"{datetime.now().strftime('%d/%m/%y %H:%M')}\n"
            )
            handle.write(shellz.init_lines.format(env_file_path=config.model_config["env_file"]))

    config.set_key(
        "aiida_venv_dir",
        os.environ.get("WORKON_HOME", config.aiida_venv_dir.as_posix()),
    )
    config.set_key("aiida_project_dir", config.aiida_project_dir.as_posix())
    print("\n✨🚀 AiiDA-project has been initialised! 🚀✨\n")
    print("[bold blue]Info:[/] For the changes to take effect, run the following command:")
    print(f"\n    source {shellz.config_file.resolve()}\n")
    print("or simply open a new terminal.")


@app.command()
def create(
    name: str,
    engine: EngineType = EngineType.venv,
    core_version: str = "latest",
    plugins: Annotated[
        List[str], typer.Option("--plugin", "-p", help="Extra plugins to install.")
    ] = [],
    python: Annotated[
        Optional[str],
        typer.Option(
            "--python",
            help="Path to the Python interpreter to use for the environment.",
        ),
    ] = None,
):
    """Create a new AiiDA project named NAME."""
    from ..config import ProjectConfig
    from ..project import ProjectDict

    config = ProjectConfig()
    if config.is_not_initialised():
        return

    venv_path = config.aiida_venv_dir / Path(name)
    project_path = config.aiida_project_dir / Path(name)

    # Temporarily block `conda` engines until we provide support again
    if engine is EngineType.conda:
        print(
            "[bold red]Error:[/bold red] The `conda` engine is currently disabled until we restore "
            "support."
        )
        return

    project = load_project_class(engine.value)(
        name=name,
        project_path=project_path,
        venv_path=venv_path,
        dir_structure=config.aiida_project_structure,
    )
    if python is None:
        python_path = Path(sys.executable)
    else:
        python_path = Path(python)
        if not python_path.exists():
            python_which = shutil.which(python)
            if python_which is None:
                python_which = shutil.which(f"python{python}")
            if python_which is None:
                print("[bold red]Error:[/bold red] Could not resolve path to Python binary.")
                return
            else:
                python_path = Path(python_which)

    print(
        "✨ Creating the project directory and environment using the Python binary:\n"
        f"   [purple]{python_path.resolve()}[/]"
    )
    project.create(python_path=python_path)

    typer.echo("🔧 Adding the AiiDA environment variables to the activate script.")
    shell = load_shell(config.aiida_project_shell)
    project.append_activate_text(shell.activate.format(env_file_path=project_path))
    project.append_deactivate_text(shell.deactivate)

    project_dict = ProjectDict()
    project_dict.add_project(project)
    print("✅ [bold green]Success:[/bold green] Project created.")

    if core_version != "latest":
        typer.echo(f"💾 Installing AiiDA core module v{core_version}.")
        project.install(f"aiida-core=={core_version}")
    else:
        typer.echo("💾 Installing the latest release of the AiiDA core module.")
        project.install("aiida-core")

    for plugin in plugins:
        if "github.com" in plugin:
            clone_path = project.project_path / Path("git") / Path(plugin.split("/")[-1]).stem
            typer.echo(f"⬇️  Cloning repo `{plugin}` from GitHub to `{clone_path.resolve()}`.")
            project.clone_repo(plugin, clone_path)
            typer.echo(f"💾 Installing local repo `{clone_path}` as editable install.")
        else:
            typer.echo(f"💾 Installing `{plugin}` from the PyPI.")
            project.install(plugin)


@app.command()
def destroy(
    name: str,
    force: Annotated[
        bool, typer.Option("--force", "-f", help="Do not ask for confirmation.")
    ] = False,
):
    """Fully remove both the virtual environment and project directory."""
    from ..config import ProjectConfig
    from ..project import ProjectDict

    if ProjectConfig().is_not_initialised():
        return

    project_dict = ProjectDict()

    try:
        project = project_dict.projects[name]
    except KeyError:
        print(f"[bold red]Error:[/bold red] No project with name {name} found!")
        return

    if not force:
        typer.confirm(
            f"❗️ Are you sure you want to delete the entire {name} project? This cannot be undone!",
            abort=True,
        )

    project.destroy()
    project_dict.remove_project(name)
    print(f"[bold green]Succes:[/bold green] Project with name {name} has been destroyed.")



# import json
# import os
# import subprocess

# import py  # type: ignore

from aiida_project.project.base import BaseProject


class CondaProject(BaseProject):
    """Conda environment."""

    _engine = "conda"


#     def __init__(self, project_path):
#         super().__init__(project_path)
#         self.conda_info = json.loads(
#             subprocess.check_output(["conda", "info", "-a", "--json"])
#         )

#     @property
#     def conda_path(self):
#         return py.path.local(self.conda_info["conda_prefix"])

#     @property
#     def venv_path(self):
#         for path in self.conda_info["envs_dirs"]:
#             env_collection_path = py.path.local(path)
#             env_path = env_collection_path.join(self.venv_name)
#             if env_path.exists():
#                 return env_path
#         return None

#     def create(self, name, python=None):
#         self._venv_name = name
#         if not self.venv_path:
#             conda_command = ["conda", "create", "-y", "-n", name, "python=2"]
#             subprocess.check_call(conda_command)

#     def add_activate_script(self, path):
#         activate_d_path = self.venv_path.join("etc", "conda", "activate.d")
#         activate_d_path.ensure_dir()
#         os.symlink(path.strpath, activate_d_path.join(path.basename).strpath)

#     def add_deactivate_scritp(self, path):
#         deactivate_d_path = self.venv_path.join("etc", "conda", "deactivate.d")
#         deactivate_d_path.ensure_dir()
#         os.symlink(path.strpath, deactivate_d_path.join(path.basename).strpath)



from pathlib import Path
from typing import Optional

import dotenv
from pydantic_settings import BaseSettings, SettingsConfigDict
from rich import print

DEFAULT_PROJECT_STRUCTURE = {
    "setup": [
        "profile",
        "computer",
        "code",
    ],
    "git": [],
}


class ProjectConfig(BaseSettings):
    """Configuration class for configuring `aiida-project`."""

    aiida_venv_dir: Path = Path(Path.home(), ".aiida_venvs")
    aiida_project_dir: Path = Path(Path.home(), "project")
    aiida_default_python_path: Optional[Path] = None
    aiida_project_structure: dict = DEFAULT_PROJECT_STRUCTURE
    aiida_project_shell: str = "bash"
    model_config = SettingsConfigDict(
        env_file=Path.home() / Path(".aiida_project.env"), env_file_encoding="utf-8"
    )

    def is_not_initialised(self):
        if dotenv.get_key(self.model_config["env_file"], "aiida_project_shell") is None:
            print("[bold red]Error:[/bold red] The AiiDA project config has not been initialised.")
            print("[bold blue]Info:[/bold blue] Please run `aiida-project init` to get started.")
            return True

    def set_key(self, key, value):
        dotenv.set_key(self.model_config["env_file"], key, value)

    def get_key(self, key):
        return dotenv.get_key(key)



"""Functionality to support various shells with `aiida-project`."""
from __future__ import annotations

from enum import Enum
from importlib import resources
from pathlib import Path

import yaml
from pydantic import BaseModel, field_validator


class ShellType(str, Enum):
    bash = "bash"
    zsh = "zsh"
    fish = "fish"


class Shell(BaseModel):
    config_file: Path
    """Path to shell configuration relative to home directory."""
    init_lines: str
    """Lines to add to the shell configuration files when using `aiida-project init`."""
    activate: str
    """AiiDA-specific lines to add to the environment's activate script."""
    deactivate: str
    """AiiDA-specific lines to add to the environment's deactivate script."""

    @field_validator("config_file")
    @classmethod
    def resolve_config_file(cls, value: Path) -> Path:
        """Resolve the shell configuration file."""
        return Path.home() / value


def load_shell(shell_str: str) -> Shell:
    """Load the project class corresponding the engine type."""
    from . import data

    with (resources.files(data) / "shell_fields.yaml").open("r") as handle:
        specs = yaml.safe_load(handle)
    return Shell(**specs[shell_str])



import shutil
import subprocess
from pathlib import Path

from aiida_project.config import ProjectConfig
from aiida_project.project.base import BaseProject


class VenvProject(BaseProject):
    """An AiiDA environment based on `venv`."""

    _engine = "venv"

    shell_activate_mapping: dict = {
        "bash": "activate",
        "zsh": "activate",
        "fish": "activate.fish",
    }
    shell_deactivate_mapping: dict = {
        "bash": "deactivate () {",
        "zsh": "deactivate () {",
        "fish": 'function deactivate -d "Exit virtual environment"',
    }

    def create(self, python_path: Path) -> None:
        super().create(python_path)
        venv_command = [f"{python_path.resolve()}", "-m", "venv", "."]
        self.venv_path.mkdir(
            exist_ok=True,
            parents=True,
        )
        subprocess.run(venv_command, cwd=self.venv_path, capture_output=True)

    def destroy(self):
        """Destroy the project."""
        super().destroy()
        shutil.rmtree(self.venv_path, ignore_errors=True)

    def append_activate_text(self, text):
        activate_file = self.shell_activate_mapping[ProjectConfig().aiida_project_shell]
        with Path(self.venv_path, "bin", activate_file).open("a") as handle:
            handle.write(text)

    def append_deactivate_text(self, text):
        activate_file = self.shell_activate_mapping[ProjectConfig().aiida_project_shell]
        with Path(self.venv_path, "bin", activate_file).open("r") as handle:
            contents = handle.read()

        # Make sure the content has the right indent - Required to satisfy Python-OCD
        text = "\n".join([" " * 4 + line.lstrip(" ") for line in text.splitlines()])
        replace_line = self.shell_deactivate_mapping[ProjectConfig().aiida_project_shell]

        with Path(self.venv_path, "bin", activate_file).open("w") as handle:
            handle.write(
                contents.replace(
                    replace_line,
                    replace_line + f"\n{text}\n",
                )
            )

    def install(self, package):
        install_command = [Path(self.venv_path, "bin", "pip").as_posix(), "install", package]
        subprocess.run(install_command, capture_output=True)

    def install_local(self, path):
        install_command = []
        install_command.append(Path(self.venv_path, "bin", "pip")).as_posix()
        install_command.extend(["install", "-e", path.as_posix()])
        subprocess.run(install_command, cwd=self.project_path)



from enum import Enum
from pathlib import Path
from typing import Dict, Type, Union

from ..config import ProjectConfig
from .base import BaseProject
from .conda import CondaProject
from .venv import VenvProject


def load_project_class(engine_type: str) -> Type[BaseProject]:
    """Load the project class corresponding the engine type."""
    engine_project_dict = {
        "venv": VenvProject,
        "conda": CondaProject,
    }
    return engine_project_dict[engine_type]


class EngineType(str, Enum):
    venv = "venv"
    conda = "conda"


class ProjectDict:
    _projects_path = Path(ProjectConfig().aiida_project_dir, ".aiida_projects")

    def __init__(self):
        if not self._projects_path.exists():
            self._projects_path.joinpath("venv").mkdir(parents=True, exist_ok=True)
            self._projects_path.joinpath("conda").mkdir(parents=True, exist_ok=True)

    @property
    def projects(self) -> Dict[str, BaseProject]:
        projects = {}
        for project_file in self._projects_path.glob("**/*.json"):
            engine = load_project_class(str(project_file.parent.name))
            project = engine.parse_file(project_file)
            projects[project.name] = project
        return projects

    def add_project(self, project: BaseProject) -> None:
        """Add a project to the configuration files."""
        with Path(self._projects_path, project.engine, f"{project.name}.json").open("w") as handle:
            handle.write(project.json())

    def remove_project(self, project: Union[str, BaseProject]) -> None:
        """Remove a project from the configuration files."""
        project = self.projects[project] if isinstance(project, str) else project
        Path(self._projects_path, project.engine, f"{project.name}.json").unlink()
